#ifndef _BIOS_H_
#define _BIOS_H_

#include <stdio.h>
#include <conio.h>
#include <string.h>
#include <ctype.h>
#include "defs.h"
#include "video.h"
#include "utils.h"

unsigned char Table[16] = {'0','1','2','3','4','5','6','7',
                           '8','9','A','B','C','D','E','F'};

#define Msg MessageBox

class TDump {
    long offset;
    unsigned CRC;
    unsigned char far *data;
    unsigned char Row, Column, Digit;
    char grid;
    bool Panel;
    void DrawFrame(unsigned char x, unsigned char y, unsigned char w, unsigned char h, unsigned char attr);
    void GetDump();
    void ShowDump(unsigned char x, unsigned char y);
    void SetCursor(unsigned char x, unsigned char y);
    void ChangeByte(unsigned char Key);
    void PrintPosition();
    void PrintValue();
  public:
    TDump();
    void Update();
    void DrawAll();
    void HandleEvent(char Key, char oldKey);
    void Save(char *filename);
    void GoTo(char *addr);
    void ToggleGrid() { grid ^= 1; }
    long Search(char *s, unsigned offs);
    long GetPosition();
};

TDump::TDump()
{
  Row = 0;
  Column = 0;
  Digit = 0;
  Panel = true;
  offset = 0;
  grid = 0;
  DrawAll();
  SetCursor(12, 2);
}

void TDump::Update()
{
  GetDump();
  ShowDump(2, 2);
  PrintPosition();
  PrintValue();
}

void TDump::PrintPosition()
{
  long offs = (offset + Row) * 16 + Column;
  unsigned char b;

  for (int i = 0; i < 8; i++)
  {
    b = (offs >> ((7 - i) * 4)) & 15;
    out_char(i + 67, 24, Table[b], 0x70);
  }
}

long TDump::GetPosition()
{
  return (offset + Row) * 16 + Column;
}

void TDump::GoTo(char *addr)
{
  int i = 0;
  long offs = 0;

  for (int i = 0; i < strlen(addr); i++)
  {
    addr[i] = toupper(addr[i]);

    if (addr[i] >= '0' && addr[i] <= '9')
      offs |= addr[i] - '0';
    if (addr[i] >= 'A' && addr[i] <= 'F')
      offs |= addr[i] - 'A' + 10;

    offs <<= 4;
  }

  offset = offs >> 8;
  GetDump();
  Row = 0;
  Column = (offs >> 4) & 15;
}

long TDump::Search(char *s, unsigned offs)
{
  long i = (offset + Row) * 16 + Column + offs;
  int j;

  if (!s[0]) return GetPosition();

  _setcursortype(_NOCURSOR);
  ClearBuf();

  for (; i < 0x100000 - strlen(s); i++)
  {
    data = (char far *)MK_FP((unsigned)(i >> 4), i & 15);
    for (j = 0; j < strlen(s); j++)
      if (toupper(data[j]) != toupper(s[j]))
      {
        j = 0;
        break;
      }
    if (j)
    {
      Column = i & 15;
      Row = 0;
      offset = i >> 4;
      _setcursortype(_NORMALCURSOR);
      return i;
    }
  }
  Msg("Error", "Specified string not found\n", 0);
  return GetPosition();
}

void TDump::PrintValue()
{
  unsigned char b = data[Row * 16 + Column];

  out_char(17, 1, Table[b >> 4], 0x1A);
  out_char(18, 1, Table[b & 15], 0x1A);

  out_char(25, 1, Table[b / 100], 0x1A);
  out_char(26, 1, Table[(b % 100) / 10], 0x1A);
  out_char(27, 1, Table[b % 10], 0x1A);

  out_char(34, 1, Table[b >> 6], 0x1A);
  out_char(35, 1, Table[(b >> 3) & 7], 0x1A);
  out_char(36, 1, Table[b & 7], 0x1A);

  for (int i = 0; i < 8; i++)
    out_char(43 + i, 1, Table[(b >> (7 - i)) & 1], 0x1A);
}

void TDump::DrawAll()
{
  out_hotstring(25, 24, "~F2~ Save  ~F5~ GoTo  ~F7~ Search  Position: 00000000 ", MENU_COLOR, HOT_COLOR);
  DrawFrame(0, 1, 80, 23, 0x1F);
  out_string(12, 1, "Hex: ** Dec: *** Oct: *** Bin: ********", 0x1E);
  out_string(68, 8, " ASCII ", 0x1E);
  fill(60, 2, 1, 21, '', 0x1F);
  fill(10, 2, 1, 21, ':', 0x1F);
  out_char(60, 1, '', 0x1F);
  out_char(60, 23, '', 0x1F);

  for (int i = 0; i < 16; i++)
  {
    out_char(i * 3 + 12, 23, Table[i >> 4], 0x1E);
    out_char(i * 3 + 13, 23, Table[i & 15], 0x1E);
    if (i < 15) out_char(i * 3 + 14, 23, '-', 0x1E);
  }

  Update();
}

void TDump::DrawFrame(unsigned char x, unsigned char y, unsigned char w, unsigned char h, unsigned char attr)
{
  int i, j;

  for (i = x + 1; i < x + w - 1; i++)
  {
    out_char(i, y, 0xCD, attr);
    out_char(i, y + h - 1, 0xCD, attr);
  }

  for (i = y + 1; i < y + h - 1; i++)
  {
    out_char(x, i, 0xBA, attr);
    out_char(x + w - 1, i, 0xBA, attr);
  }

  out_char(x, y, 0xC9, attr);
  out_char(x + w - 1, y, 0xBB, attr);
  out_char(x, y + h - 1, 0xC8, attr);
  out_char(x + w - 1, y + h - 1, 0xBC, attr);

  for (i = x + 1; i < x + w - 1; i++)
    for (j = y + 1; j < y + h - 1; j++)
      out_char(i, j, 0x20, attr);
}

void TDump::GetDump()
{
  data = (char far *)MK_FP(offset, 0x0000);
}

void TDump::ShowDump(unsigned char x, unsigned char y)
{
  int i, j = 0;
  unsigned char b, attr;
  long offs = offset * 16;

  for (; j < 21; j++, offs += 16)
  {
    for (i = 0; i < 8; i++)
    {
      b = (offs >> ((7 - i) * 4)) & 15;
      out_char(x + i, y + j, Table[b], 0x1F);
    }
    for (i = 0; i < 16; i++)
    {
      b = data[i + j * 16];
      if ((grid && (i == Column || j == Row)) || (i == Column && j == Row))
        attr = 0x0B;
      else
        attr = 0x1B;
      out_char(i * 3 + x + 10, j + y, Table[b >> 4], attr);
      out_char(i * 3 + x + 11, j + y, Table[b & 15], attr);
      out_char(i + x + 60, j + y, b, attr);
    }
  }
}

void TDump::SetCursor(unsigned char x, unsigned char y)
{
  if (Panel)
    gotoxy(Column * 3 + Digit + x + 1, Row + y + 1);
  else
    gotoxy(Column + x + 51, Row + y + 1);
}

void TDump::ChangeByte(unsigned char Key)
{
  unsigned char a, b;

  if (Panel)
  {
    b = data[Row * 16 + Column];

    if (Key >= '0' && Key <= '9')
      a = Key - '0';
    if (Key >= 'A' && Key <= 'F')
      a = Key - 'A' + 10;
    if (Key >= 'a' && Key <= 'f')
      a = Key - 'a' + 10;

    if (!Digit)
      b = (b & 0x0F) | (a << 4);
    else
      b = (b & 0xF0) | a;

    data[Row * 16 + Column] = b;
  }
  else
    data[Row * 16 + Column] = Key;
}

void TDump::Save(char *filename)
{
  FILE *f;
  char name[15];
  char msg[60] = "(cannot write file ";

  if (!filename[0])
    strcpy(name, "memory.dat");
  else
    strcpy(name, filename);

  f = fopen(name, "wb");
  if (f)
  {
    for (int i = 0; i < 32; i++)
    {
      data = (char far *)MK_FP(i * 0x800, 0x0000);
      for (int j = 0; j < 0x8000; j++)
        putc(data[j], f);
    }
    data = (char far *)MK_FP(offset, 0x0000);
    fclose(f);
    Msg("Information", "Data from the system memory saved\n", 0);
  }
  else
  {
    strcat(msg, name);
    strcat(msg, ")\n");
    Msg("Error", msg, 0);
  }
}

void TDump::HandleEvent(char Key, char oldKey)
{
  if (!oldKey)
  {
    switch (Key)
    {
      case 77:
a:             if (Column < 15 || (!Digit && Panel))
               {
                 if (Digit || !Panel)
                   Column++;
                 if (Panel)
                   Digit ^= 1;
                 break;
               }
               Column = 0;
               if (Panel) Digit = 0;
      case 80: if (Row < 20)
                 Row++;
               else
                 offset++;
               break;
      case 75: if (Column || (Digit && Panel))
               {
                 if (!Digit || !Panel)
                   Column--;
                 if (Panel)
                   Digit ^= 1;
                 break;
               }
               Column = 15;
               if (Panel) Digit = 1;
      case 72: if (Row)
                 Row--;
               else
                 offset--;
               break;
      case 81: offset += 21; break;
      case 73: offset -= 21; break;
      case 71: Column = 0; break;
      case 79: Column = 15; break;
    }
    if (offset < 0)
    {
      offset = 0;
      Column = 0;
      if (Panel) Digit = 0;
      Row = 0;
    }
    if (offset > 65515L)
    {
      offset = 65515L;
      Column = 15;
      if (Panel) Digit = 1;
      Row = 20;
    }
  }
  else
  {
    if (Key == 9)
      Panel = !Panel;
    else
      if (Key && Key != 27)
      {
        ChangeByte(Key);
        goto a;
      }
  }
  SetCursor(12, 2);
}

#endif
